// TypeScript Version: 3.0

declare module 'react-context-toolbox/forwardRef' {
  import * as React from 'react';

  interface ForwardRefOptions<TProps> {
    displayName?: string;
    propTypes?: React.ValidationMap<TProps>;
    defaultProps?: Partial<TProps>;
    allowFallback?: boolean;
  }

  function forwardRef<TRef, TProps>(
    renderFn: (
      props: TProps & { children?: React.ReactNode },
      ref: React.Ref<TRef> | null,
    ) => React.ReactElement<any> | null,
    options?: ForwardRefOptions<TProps>,
  ): React.ForwardRefExoticComponent<
    React.PropsWithRef<TProps> & React.RefAttributes<TRef>
  >;

  export default forwardRef;
}

declare module 'react-context-toolbox/mapContextToProps' {
  import * as React from 'react';

  type ElementType = React.ComponentType<any> | keyof JSX.IntrinsicElements;

  type Omit<T, U> = Pick<T, Exclude<keyof T, keyof U>>;

  type GetProps<C> = C extends React.ComponentType<infer P> ? P : never;

  interface ContextInjectedComponent<TComponent, TInjectedProps, TExtraProps>
    extends React.ForwardRefExoticComponent<
        Omit<GetProps<TComponent>, TInjectedProps> & TExtraProps
      > {}

  // Single Context
  function mapContextToProps<TComponent, TContext, TContextProps, TOwnProps>(
    context: React.Context<TContext> | [React.Context<TContext>],
    mapToProps: (ctxValue: TContext, props: TOwnProps) => TContextProps,
    Component: TComponent,
  ): ContextInjectedComponent<TComponent, TContextProps, TOwnProps>;
  function mapContextToProps<TContext, TContextProps, TOwnProps>(
    context: React.Context<TContext> | [React.Context<TContext>],
    mapToProps: (ctxValue: TContext, props: TOwnProps) => TContextProps,
  ): <TComponent>(
    component: TComponent,
  ) => ContextInjectedComponent<TComponent, TContextProps, TOwnProps>;

  // 2 Contexts
  function mapContextToProps<
    TComponent,
    TContext1,
    TContext2,
    TContextProps,
    TOwnProps
  >(
    context: [React.Context<TContext1>, React.Context<TContext2>],
    mapToProps: (
      c1: TContext1,
      c2: TContext2,
      props: TOwnProps,
    ) => TContextProps,
    Component: TComponent,
  ): ContextInjectedComponent<TComponent, TContextProps, TOwnProps>;
  function mapContextToProps<TContext1, TContext2, TContextProps, TOwnProps>(
    context: [React.Context<TContext1>, React.Context<TContext2>],
    mapToProps: (
      c1: TContext1,
      c2: TContext2,
      props: TOwnProps,
    ) => TContextProps,
  ): <TComponent>(
    component: TComponent,
  ) => ContextInjectedComponent<TComponent, TContextProps, TOwnProps>;

  // 3 Contexts
  function mapContextToProps<
    TComponent,
    TContext1,
    TContext2,
    TContext3,
    TContextProps,
    TOwnProps
  >(
    context: [
      React.Context<TContext1>,
      React.Context<TContext2>,
      React.Context<TContext3>
    ],
    mapToProps: (
      c1: TContext1,
      c2: TContext2,
      c3: TContext3,
      props: TOwnProps,
    ) => TContextProps,
    Component: TComponent,
  ): ContextInjectedComponent<TComponent, TContextProps, TOwnProps>;
  function mapContextToProps<
    TContext1,
    TContext2,
    TContext3,
    TContextProps,
    TOwnProps
  >(
    context: [
      React.Context<TContext1>,
      React.Context<TContext2>,
      React.Context<TContext3>
    ],
    mapToProps: (
      c1: TContext1,
      c2: TContext2,
      c3: TContext3,
      props: TOwnProps,
    ) => TContextProps,
  ): <TComponent>(
    component: TComponent,
  ) => ContextInjectedComponent<TComponent, TContextProps, TOwnProps>;

  // 4 Contexts
  function mapContextToProps<
    TComponent,
    TContext1,
    TContext2,
    TContext3,
    TContext4,
    TContextProps,
    TOwnProps
  >(
    context: [
      React.Context<TContext1>,
      React.Context<TContext2>,
      React.Context<TContext3>,
      React.Context<TContext4>
    ],
    mapToProps: (
      c1: TContext1,
      c2: TContext2,
      c3: TContext3,
      c4: TContext4,
      props: TOwnProps,
    ) => TContextProps,
    Component: TComponent,
  ): ContextInjectedComponent<TComponent, TContextProps, TOwnProps>;
  function mapContextToProps<
    TContext1,
    TContext2,
    TContext3,
    TContext4,
    TContextProps,
    TOwnProps
  >(
    context: [
      React.Context<TContext1>,
      React.Context<TContext2>,
      React.Context<TContext3>,
      React.Context<TContext4>
    ],
    mapToProps: (
      c1: TContext1,
      c2: TContext2,
      c3: TContext3,
      c4: TContext4,
      props: TOwnProps,
    ) => TContextProps,
  ): <TComponent>(
    component: TComponent,
  ) => ContextInjectedComponent<TComponent, TContextProps, TOwnProps>;

  export default mapContextToProps;
}
